/* 
 * Copyright (C) 2000 plitsch-platsch
 * 
 * xine_dvd_plugin is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine_dvd_plugin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 */

#include <dlfcn.h>
#include <stdio.h>
#include <fcntl.h>
#if defined(__NetBSD__) || defined(__OpenBSD__) || defined(__FreeBSD__)
# include <sys/dvdio.h>
# include <sys/cdio.h> /* CDIOCALLOW etc... */
#elif defined(__linux__)
# include <linux/cdrom.h>
#else
# error "Need the DVD ioctls"
#endif
#include <sys/ioctl.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>

#include "input_plugin.h"
#include "dvd_udf.h"
#include "css.h"
#include "ifo.h"
#include "bswap.h"

#define DVD     "/dev/dvd"
#define RDVD    "/dev/rdvd"
/* 
 * Define this and input_plugin_get_autoplay_list() 
 * will return all of titles/chapters
 */
//#define ALL_TITLES 1


/*
 * global Variables:
 */

int                bEncrypted;
char               key_disk [2048];
char               key_title [5];
off_t              file_size, file_size_left;
off_t              file_lbstart, file_lbcur;
int                dvd_fd, raw_fd;
int                gVTSMinor, gVTSMajor;

/*
 * ifo stuff
 */
ifo_t               *ifo;
off_t                title_start;
pgc_t               *pgci;          
ifo_pgci_caddr_t    *cpli;          
ifo_pgc_cpos_t      *cpos;          
int32_t              cpos_index;    
ifo_caddr_t         *caddr;         
int32_t              caddr_index;    
uint32_t             caddr_index_max;
uint32_t             cellp_pos;
uint32_t             cell_end; 
off_t                byte_pos;
uint8_t             *clink;
int                  gChapter, gTitle, gToChapter;
uint32_t             lb_end; /* lb to stop at (==start lb of next chapter) */
ifo_ptt_t           *ptt;


/*
 * udf dir function 
 */

#define MAX_DIR_ENTRIES 1024

char              *filelist[MAX_DIR_ENTRIES];
char              *filelist2[MAX_DIR_ENTRIES];
int                nFiles;

#define MAX(a,b) ( (a) > (b) ? (a) : (b) )
#define MIN(a,b) ( (a) < (b) ? (a) : (b) )

static void reset_dvd_drive () {
  int fd;
#ifndef __FreeBSD__
  dvd_authinfo ai;
#else
  struct dvd_authinfo ai;
#endif

  int i;

  /*
   * reset drive 
   */

  fd = open(DVD, O_RDONLY | O_NONBLOCK);

  if (fd < 0) {
    printf ("input_dvd: reset_dvd_drive: unable to open dvd drive (%s): %s\n", DVD, strerror(errno));
    return ;
  }

  for (i = 0; i < 4; i++) {
    memset(&ai, 0, sizeof(ai));
#ifndef __FreeBSD__
    ai.type = DVD_INVALIDATE_AGID;
    ai.lsa.agid = i;
    ioctl(fd, DVD_AUTH, &ai);
#else
    ai.format = DVD_INVALIDATE_AGID;
    ai.agid = i;
    ioctl(fd, DVDIOCREPORTKEY, &ai);
#endif
  }

  close (fd); 
}

static int openDrive () {
  
  dvd_fd = open(DVD, O_RDONLY | O_NONBLOCK);

  if (dvd_fd < 0) {
    printf ("input_dvd: unable to open dvd drive (%s): %s\n", DVD, strerror(errno));
    return -1;
  }

  raw_fd = open(RDVD, O_RDONLY | O_NONBLOCK);
  if (raw_fd < 0) {
    raw_fd = dvd_fd;
  }
  return raw_fd;
}

static void closeDrive () {

  if (dvd_fd<0)
    return;

  close (dvd_fd);
  if (raw_fd!=dvd_fd)
    close (raw_fd);

  dvd_fd = -1;
}

static void unlockDrive (off_t lbnum) {

  lseek (dvd_fd, lbnum * (off_t) DVD_VIDEO_LB_LEN, SEEK_SET) ;

  /* printf ("seek to %d: %d\n",lbnum, ret); */
  
  CSSAuthDisc (dvd_fd, key_disk);
  
  closeDrive ();
  
  openDrive ();
  
  lseek (dvd_fd, lbnum * (off_t) DVD_VIDEO_LB_LEN, SEEK_SET) ;

  if (CSSAuthTitle (dvd_fd, key_title, lbnum)>=0) {
    
    if (CSSDecryptTitleKey (key_title, key_disk) >= 0) {
      
      bEncrypted = 1;
    } else {
      perror ("Decrypting title key");
      
      bEncrypted = 0;
    }
    
  } else {
    bEncrypted = 0;
    perror ("Authenticate title");
  }
}


/*
 * try to open dvd, authenticate and prepare to read/decrypt >filename<
 *
 * returns lbnum on success, 0 otherwise
 */

static int openDVDFile (char *filename, off_t *size) {

#ifndef __FreeBSD__
  dvd_struct         dvd;
#else
  struct dvd_struct         dvd;
#endif

  char               str[256];
  off_t              lbnum;

  printf ("input_dvd : openDVDFile >%s<\n",filename);

  reset_dvd_drive ();

  
  if (openDrive() < 0) 
    return 0;

  /*
   * encrypted ?
   */

#ifndef __FreeBSD__
  dvd.copyright.type = DVD_STRUCT_COPYRIGHT;
  dvd.copyright.layer_num=0;

  if (ioctl (dvd_fd, DVD_READ_STRUCT, &dvd) < 0) {
    perror ("DVD_READ_STRUCT");
    return 0;
  }

  bEncrypted = (dvd.copyright.cpst != 0) ;

#else
  dvd.format    = DVD_STRUCT_COPYRIGHT;
  dvd.layer_num = 0;
  
  if (ioctl(dvd_fd, DVDIOCREADSTRUCTURE, &dvd) < 0) {
    fprintf (stderr, "Could not read Copyright Struct\n");
    return 0;
  }

  bEncrypted = (dvd.cpst != 0);
#endif

  closeDrive ();

  if (bEncrypted)
    printf ("encrypted\n");
  else
    printf ("unencrypted\n");

  openDrive ();

  sprintf (str, "/VIDEO_TS/%s", filename);

  printf ("UDFFindFile %s\n",str);
    
  if ((lbnum=UDFFindFile(dvd_fd, str, size))) {

    unlockDrive (lbnum);

    lseek (raw_fd, lbnum * (off_t) DVD_VIDEO_LB_LEN, SEEK_SET) ;

  } else {
    printf ("input_dvd: cannot open file >%s<\n", filename);

    closeDrive ();

    return 0;
  }

  return lbnum;
}


static void input_plugin_init () {
  int i, fd;

  ifo = NULL;
  
  /*
   * allocate space for directory listing
   */

  for (i=0; i<MAX_DIR_ENTRIES; i++) {
    filelist[i] = (char *) malloc (256);
    filelist2[i] = (char *) malloc (256);
  }

  /*
   * find out if raw device is available
   */

  fd = open(RDVD, O_RDONLY | O_NONBLOCK);
  if (fd < 0) {
    printf ("input_dvd: unable to open raw dvd drive (%s): %s\n", RDVD, strerror(errno));

  } else {
    printf ("input_dvd: using raw device %s\n", RDVD);
    close (fd);
  }
}

static int _dvd_next_cell ()
{
  
  int old_caddr_index = caddr_index;

  /*
  printf( "looking for cell %x:%x (%x/%x)\n", 
	  cpos[cpos_index].vob_id, 
	  cpos[cpos_index].cell_id, 
	  cpos_index, caddr_index_max);
  */

  if (cpos_index > caddr_index_max) {
    printf("input_dvd: cpos_index >caddr_index_max => finish\n");
    return -1;
  }

  while (++caddr_index < caddr_index_max) {

    if ((cpos[cpos_index].vob_id == caddr[caddr_index].vob_id)
	&& (cpos[cpos_index].cell_id == caddr[caddr_index].cell_id)) {
      goto cell_found;
    }
  }

  caddr_index = old_caddr_index;

  return -1;

cell_found:

  /* for ILVUs which don't end at cell boundary */
  cellp_pos = MAX (be2me_32 (caddr[caddr_index].start), 
		   be2me_32 (cpli[cpos_index].vobu_start));
  cell_end = MIN (be2me_32 (caddr[caddr_index].end), 
		  be2me_32 (cpli[cpos_index].vobu_last_end));

  /* printf("|%x %x|", cellp_pos, cell_end); */
  return 0;
}

static int parseIFOs () {

  int      found_num, found_sub, num, subnum;
  off_t    size;
  int      i;

  /* find right IFO first */

  UDFListDir (dvd_fd, "/VIDEO_TS", MAX_DIR_ENTRIES, filelist, &nFiles);

  // UDFListDir set nFiles to -1 if UDFFindPartition() failed */
  if(nFiles < 0)
    return 0;
  
  title_start = 0;
  found_num = -1; found_sub = -1;
  for (i=0; i<nFiles; i++) {
    if (sscanf (filelist[i], "VTS_%d_%d.VOB", &num, &subnum)==2) {
      if (subnum>found_sub) {
	found_sub = subnum;
	found_num = num;
      }
    }
    
  }

  if (found_num>-1) {
    char str[1024];
    sprintf (str, "VIDEO_TS/VTS_%02d_0.IFO", found_num);
    
    title_start = UDFFindFile (dvd_fd, str, &size);
    if (title_start)
      printf ("Using ifo VTS_%02d_0.IFO\n",found_num);
  } 
  
  if (!title_start) {

    title_start = UDFFindFile (dvd_fd, "/VIDEO_TS/VIDEO_TS.IFO", &size);
    
    printf ("first ifo at lba %d\n",title_start);
    
    ifo = ifoOpen (dvd_fd, title_start * 2048);
    
    if (!ifo)
      printf ("ifoOpen failed\n");
    
    title_start += ifoGetTSPoffset (ifo, 0);
    
    ifoClose (ifo);
    
    printf ("second ifo at lba %d\n",title_start);
  }


  ifo = ifoOpen (dvd_fd, title_start * 2048);
  
  if (!ifo) {
    printf ("ifoOpen 2 failed\n");
    return 0;
  }

  caddr = ifoGetCellAddr (ifo);
  title_start += (off_t) ifoGetVOBStart (ifo);

  ptt = ifoGetPTT (ifo);

  unlockDrive (title_start);

  return 1;
}

static int openDVDChapter (int title_num, int chapter, int to_chapter) {

  ifo_ptt_sub_t *title;

  gTitle     = title_num;
  gChapter   = chapter;
  gToChapter = to_chapter;

  reset_dvd_drive ();

  if (openDrive()<0) 
    return 0;
  
  /*
   * parse IFO files
   */

  if (!parseIFOs ())
    return 0;

  title = ptt->title+title_num;

  /*
   * use cell link to find first cell of selected chapter 
   */

  pgci = ifoGetPGCI (ifo, IFO_TITLE, title_num);
  
  if (!pgci) {
    printf ("error get PGCI\n");
    return 0;
  }

  clink = ifoGetProgramMap (pgci);

  caddr_index_max = ifoGetCellAddrNum (ifo, IFO_TITLE);

  /* Cell Position */
  if (!(cpos = ifoGetCellPos (pgci))) {
    printf ("error ifoGetCellPos");
    return 0;
  }

  /* Cell Play Information */
  if (!(cpli = ifoGetCellPlayInfo (pgci))) {
    printf("error ifoGetCellPlayInfo");
    return 0;
  }

  /*
   * find out start lb of next chapter (if there is one)
   */

  if (to_chapter < (title->num-1)) {
    cpos_index     = clink[to_chapter+1]-1;
    caddr_index    = -1;
    cellp_pos      = 0;
    cell_end       = 0;

    if (++cellp_pos > cell_end) {
      if (_dvd_next_cell () < 0) {
	cpos_index++;
	_dvd_next_cell ();
      }
    }

    lb_end = title_start + cellp_pos;

  } else {
    lb_end = 0;
  }

  /*
   * calculate length of input
   * (I know this is pathetic, but it works! ;> )
   */

  cpos_index     = clink[chapter]-1;
  caddr_index    = -1;
  cellp_pos      = 0;
  cell_end       = 0;
  
  file_size = 0;

  while (1) {
    if ((cellp_pos+title_start)==lb_end) /* end reached ? */
      break;
    if (++cellp_pos > cell_end) {       /* next cell */
      if (_dvd_next_cell () < 0) {
	cpos_index++;
	if (_dvd_next_cell () < 0)
	  break;
      }
    }
    file_size += DVD_VIDEO_LB_LEN;
  }

  file_size_left = file_size;

  printf ("input length : %lld\n",file_size);

  /* reset values after length calculation */

  cpos_index     = clink[chapter]-1;

  caddr_index = -1;
  cellp_pos = 0;
  cell_end = 0;
  byte_pos = 0;

  return 1;
}

static int input_plugin_open (char *mrl) {

  char *filename;
  int title, chapter, to_chapter;

  printf ("input dvd : input_plugin_open >%s<\n",mrl);

  /*
   * do we handle this kind of MRL ?
   */

  if (strncmp (mrl, "dvd://",6))
    return 0;

  filename = &mrl[6];

  printf ("input dvd : input_plugin_open media type correct. file name is %s\n",
	  filename);

  if (sscanf (filename, "t%dc%dt%d", &title, &chapter, &to_chapter) == 3) {

    printf ("IFO-mode playing title %d from chapter %d -> chapter %d\n",
	    title, chapter, to_chapter);

    return openDVDChapter (title, chapter, to_chapter);

  } else {

    sscanf (filename, "VTS_%d_%d.VOB", &gVTSMajor, &gVTSMinor);

    ifo = NULL;

    file_lbstart = openDVDFile (filename, &file_size) ;
    file_lbcur = file_lbstart;

    printf ("opened - file size : %ld\n", file_size);

    if (!file_lbstart) {
      printf ("unable to find >%s< on dvd.\n",filename);
      return 0;
    }

    file_size_left = file_size;
  }
 
  return 1 ;
}


static uint32_t input_plugin_read (char *buf, uint32_t nlen) {

  if (nlen != DVD_VIDEO_LB_LEN) {
    printf ("ERROR in input_dvd plugin read: %d bytes is not a sector!\n",nlen);
    return 0;
  }

  if (ifo) {

    if ((cellp_pos+title_start)==lb_end)
      return 0;

    if (++cellp_pos > cell_end) {       /* next cell */
      if (_dvd_next_cell () < 0) {
	cpos_index++;
	if (_dvd_next_cell () < 0)
	  return 0;
      }
    }

    byte_pos += DVD_VIDEO_LB_LEN;
    lseek (raw_fd, (title_start + cellp_pos) * DVD_VIDEO_LB_LEN, SEEK_SET) ;

  } else {

    if (file_size_left < nlen)
      return 0;
    
    file_lbcur++;
    file_size_left -= DVD_VIDEO_LB_LEN;

  }
    
  if (read (raw_fd, buf, DVD_VIDEO_LB_LEN)) {

    /* printf ("read to %08x => %02x %02x %02x %02x\n",buf,buf[0],buf[1],buf[2],buf[3]); */
    
    if (bEncrypted)
      CSSDescramble (buf, key_title);
    
    
    return DVD_VIDEO_LB_LEN;
  } else
    printf ("read error in input_dvd plugin\n");

  return 0;
}

static off_t input_plugin_seek (off_t offset, int origin) {

  if (ifo) {
    off_t seek_pos;

    switch (origin) {
    case SEEK_END:
      seek_pos = file_size - offset;

    case SEEK_SET:
      seek_pos = offset;
      break;

    case SEEK_CUR:

      if (offset==0) {
	return byte_pos;
      }

      seek_pos = byte_pos + offset;
      break;
    default:
      printf ("error in input dvd plugin seek:%d is an unknown origin\n",origin);
    }

    cpos_index     = clink[gChapter]-1;
    caddr_index = -1;
    cellp_pos = 0;
    cell_end = 0;
    byte_pos = 0;

    while (byte_pos < seek_pos) {
      if (++cellp_pos > cell_end) {       /* next cell */
	if (_dvd_next_cell () < 0) {
	  cpos_index++;
	  if (_dvd_next_cell () < 0)
	    break;
	}
      }
      byte_pos += DVD_VIDEO_LB_LEN;
    }

    return byte_pos;
  } else {

    offset /= DVD_VIDEO_LB_LEN;

    switch (origin) {
    case SEEK_END:
      offset = (file_size / DVD_VIDEO_LB_LEN) - offset;

    case SEEK_SET:
      file_lbcur = file_lbstart + offset;
      file_size_left = file_size - (offset * DVD_VIDEO_LB_LEN);
      break;
    case SEEK_CUR:

      if (offset) {
	file_lbcur += offset;
	file_size_left = file_size - ((file_lbcur - file_lbstart) * DVD_VIDEO_LB_LEN);
      } else {
	return (file_lbcur - file_lbstart) * (off_t) DVD_VIDEO_LB_LEN;
      }
      break;
    default:
      printf ("error in input dvd plugin seek:%d is an unknown origin\n",origin);
    }

    return lseek (raw_fd, file_lbcur * (off_t) DVD_VIDEO_LB_LEN, SEEK_SET) - file_lbstart * (off_t) DVD_VIDEO_LB_LEN;
  }
  return 0;
}

static off_t input_plugin_get_length () {

  return file_size;
}

static uint32_t input_plugin_get_capabilities () {
  return INPUT_CAP_SEEKABLE | INPUT_CAP_BLOCK | INPUT_CAP_AUTOPLAY;
}

static uint32_t input_plugin_get_blocksize () {
  return DVD_VIDEO_LB_LEN;
}

static int input_plugin_eject (void) {
  int ret, status, fd;

  if((fd = open(DVD, O_RDONLY|O_NONBLOCK)) > -1) {

#if defined (__linux__)
    if((status = ioctl(fd, CDROM_DRIVE_STATUS, CDSL_CURRENT)) > 0) {
      switch(status) {
      case CDS_TRAY_OPEN:
	if((ret = ioctl(fd, CDROMCLOSETRAY)) != 0) {
	  printf("CDROMCLOSETRAY failed: %s\n", strerror(errno));  
	}
	break;
      case CDS_DISC_OK:
	if((ret = ioctl(fd, CDROMEJECT)) != 0) {
	  printf("CDROMEJECT failed: %s\n", strerror(errno));  
	}
	break;
      }
    }
    else {
      printf("CDROM_DRIVE_STATUS failed: %s\n", 
	     strerror(errno));
      close(fd);
      return 0;
    }

#elif defined (__NextBSD__) || defined (__OpenBSD__) || defined (__FreeBSD__)

    if (ioctl(fd, CDIOCALLOW) == -1) {
      perror("ioctl(cdromallow)");
    } else {
      if (ioctl(fd, CDIOCEJECT) == -1) {
        perror("ioctl(cdromeject)");
      }
    }

#endif

    close(fd);
  }

  return 1;
}

static void input_plugin_close () {
  closeDrive();

  if (ifo) {
    ifoClose (ifo);
    ifo = NULL;
  }
}

static char *input_plugin_get_identifier () {
  return "DVD";
}

static char** input_plugin_get_dir (char *filename, int *nEntries) {

  int i, fd;

  if (filename) {
    *nEntries = 0;
    return NULL;
  }

  if((fd = open(DVD, O_RDONLY|O_NONBLOCK)) > -1) {

    int    nFiles, nFiles2;

    UDFListDir (fd, "/VIDEO_TS", MAX_DIR_ENTRIES, filelist, &nFiles);

    nFiles2 = 0;
    for (i=0; i<nFiles; i++) {
      int nLen;

      nLen = strlen (filelist[i]);

      if (nLen<4) 
	continue;
      
      if (!strcasecmp (&filelist[i][nLen-4], ".VOB")) {

	sprintf (filelist2[nFiles2], "dvd://%s",filelist[i]); 

	nFiles2++;
      }

    }

    *nEntries = nFiles2;

    close (fd);

  } else {
    *nEntries = 0;
    return NULL;
  }

  return filelist2;
}

static char **input_plugin_get_autoplay_list (int *nFiles) {

  int            i, j=0, alltitles, nfilestmp;
  ifo_ptt_t     *ptt;
  ifo_ptt_sub_t *title;
  static char *fl[MAX_DIR_ENTRIES];
  char buf[256]="";

  printf ("input_dvd: get_autoplay_list\n"); 

  reset_dvd_drive ();

  if (openDrive()<0) {
    *nFiles = 0;
    return 0;
  }
  
  /*
   * parse IFO files
   */

  if (!parseIFOs ()) {
    *nFiles = 0;
    return NULL;
  }

  ptt = ifoGetPTT (ifo);

  title = ptt->title; /* assuming title 0 is the movie */
  alltitles = ptt->num;
  nfilestmp = 0;

#ifdef ALL_TITLES
  for(j = 0; j < alltitles; j++) {
#endif
    for (i = 0; i < title->num; i++, nfilestmp++) {
      sprintf (buf, "dvd://t%dc%dt%d", j, i, i);
      if((fl[nfilestmp] = (char *) realloc(fl[nfilestmp], 
					   strlen(buf)+1)) != NULL) {
  	strcpy(fl[nfilestmp], buf);
	//	fprintf(stderr, "%s ", fl[nfilestmp]);
      }
      else
  	fprintf(stderr, "%s(%d): realloc() failed: %s\n",
  		__FILE__, __LINE__, strerror(errno));
    }
#ifdef ALL_TITLES    
    title++;
  }
#endif

  fl[nfilestmp] = NULL;
  *nFiles = nfilestmp;

  closeDrive ();

  return fl;
}

static int input_plugin_is_branch_possible (char *next_mrl) {
  
  char *filename;
  int   title, chapter, to_chapter;

  printf ("input_dvd: is_branch_possible to %s ?\n", next_mrl);

  /*
   * do we handle this kind of MRL ?
   */

  if (strncmp (next_mrl, "dvd://",6))
    return 0;

  filename = &next_mrl[6];

  if (sscanf (filename, "t%dc%dt%d", &title, &chapter, &to_chapter) == 3) {

    if ((gTitle == title) && (gToChapter == (chapter-1))) {
      printf ("input_dvd: branching is possible\n");
      return 1;
    } else 
      return 0;

  } else {
    int vts_minor, vts_major;
    if (sscanf (filename, "VTS_%d_%d.VOB", &vts_major, &vts_minor) == 2) {
      if ((vts_major==gVTSMajor) && (vts_minor==(gVTSMinor+1))) {
	printf ("input_dvd: branching is possible\n");
	return 1;
      }
    }

  }

  return 0;
}

clut_t* input_plugin_get_clut(void) {

  return ifoGetCLUT(pgci);
}

void input_plugin_get_video_info(video_info_t *vinfo) {

  const ifo_video_info_t *ivinfo = ifoGetVideoInfo(ifo);
  if (ivinfo) {
    vinfo->framerate = ivinfo->system;
    vinfo->ratio = ivinfo->ratio;
    vinfo->letterboxed = ivinfo->letterboxed;
  }
  else {
    printf("no video info found\n");
  }
}

static input_plugin_t plugin_op = {
  NULL,
  NULL,
  input_plugin_init,
  input_plugin_open,
  input_plugin_read,
  input_plugin_seek,
  input_plugin_get_length,
  input_plugin_get_capabilities,
  input_plugin_get_dir,
  input_plugin_get_blocksize,
  input_plugin_eject,
  input_plugin_close,
  input_plugin_get_identifier,
  input_plugin_get_autoplay_list,
  input_plugin_is_branch_possible,
  input_plugin_get_clut,
  input_plugin_get_video_info
};

input_plugin_t *input_plugin_getinfo(uint32_t dbglvl) {

  return &plugin_op;
}
