/* 
 * Copyright (C) 2000 the xine project
 * 
 * This file is part of xine, a unix video player.
 * 
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: input_plugin.h,v 1.15 2001/01/28 02:56:06 f1rmb Exp $
 */

#ifndef HAVE_INPUT_PLUGIN_H
#define HAVE_INPUT_PLUGIN_H

#include <inttypes.h>
#include <sys/types.h>

#define INPUT_PLUGIN_MAX    127
 
#ifndef CLUT_T
#define CLUT_T
typedef struct {         /* CLUT == Color LookUp Table */
	uint8_t foo		: 8; /* UNKNOWN: 0x00? */
	uint8_t y		: 8;
	uint8_t cr		: 8;
	uint8_t cb		: 8;
} __attribute__ ((packed)) clut_t;
#endif

typedef struct {
  uint8_t framerate:2;
  uint8_t ratio:2;
  uint8_t letterboxed:1;
} video_info_t;

static int framerate_pts[] = {
  3003, /* NTSC 29.97 fps  */
  3000, /* NTSC 30 fps     */
  3600, /* PAL-B, D, G, H I, 25 fps */
  3003, /* PAL-M 29.97 fps */
  3000, /* PAL-M 30 fps    */
  3600, /* PAL-N 25 fps    */
  3600, /* PAL-Nc 25 fps   */
  3000, /* PAL 30 fps      */
  0
};

typedef struct input_plugin_s
{

  /*
   * handle of plugin
   */
  void *handle;

  /*
   * filename of plugin
   */
  char *filename;
 
  /*
   * init plugin
   */
  void (*init) (void);

  /*
   * open input MRL - return 1 if succ
   */

  int (*open) (char *mrl);

  /*
   * read nlen bytes, return number of bytes read
   */
  uint32_t (*read)(char *buf, uint32_t nlen);

  /*
   * seek position, return new position (=> way to determine cur pos)
   */
  off_t (*seek) (off_t offset, int origin);

  /*
   * return length of input (-1 => unlimited, e.g. stream)
   */

  off_t (*get_length) (void);

  /*
   * return capabilities of input source
   */

  uint32_t (*get_capabilities) (void);

  /*
   * ls function
   * return value: NULL => filename is a file, **char=> filename is a dir
   */

  char** (*get_dir) (char *filename, int *nFiles);

  /*
   * return block size of input source (if supported, 0 otherwise)
   */

  uint32_t (*get_blocksize) (void);
 
  /*
   * eject/load the media (when it's possible)
   */
  int (*eject_media) (void);

  /*
   * close input source
   */
  
  void (*close) (void);

  /*
   * return human readable identifier for this plugin
   */

  char* (*get_identifier) (void);

  /*
   * generate autoplay list
   * return value: list of MRLs
   */

  char** (*get_autoplay_list) (int *nFiles);

  /*
   * check if it is possible/valid to directly branch to this MRL
   */
  
  int (*is_branch_possible) (char *next_mrl);

  /*
   * gets the subtitle/menu palette
   */

  clut_t* (*get_clut) (void);

  /*
   * get some video info
   */

  void (*get_video_info) (video_info_t *vinfo);

} input_plugin_t;

#define INPUT_CAP_SEEKABLE 1
#define INPUT_CAP_BLOCK    2
#define INPUT_CAP_AUTOPLAY 4

#endif
